"""
This file is part of WSQL-SDK

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__author__ = "@bg"

import fnmatch
import os
import warnings
from .grammar import MacrosTokenizer


class Translator(MacrosTokenizer):
    def __init__(self, output, close_output=False):
        super().__init__()
        self.output = output
        self.close_output = close_output
        self.includes = set()
        self.workdir = os.curdir

    def write(self, raw):
        line = raw.strip('\n')
        if line:
            self.output.write(line)
            if raw.endswith('\n'):
                self.output.write('\n')

    def on_function(self, ast, body, args):
        start = 0
        for t in ast:
            self.write(body[start:t[1]])
            self.write(args[t[0].getName()])
            start = t[2]

        self.write(body[start:])

    def on_constant(self, name, value):
        self.write('-- CONSTANT {0} {1}\n'.format(name, value))

    def on_variable(self, name, value):
        self.write(value)

    def on_include(self, filename):
        filename = os.path.join(self.workdir, filename)
        dirname = os.path.dirname(filename)
        before = len(self.includes)
        for fname in fnmatch.filter(os.listdir(dirname), os.path.basename(filename)):
            self.include_file(os.path.join(dirname, fname))
        if before == len(self.includes):
            warnings.warn("Not included: %s" % filename)

    def nop(self, text):
        self.write(text)

    def include_file(self, filename):
        if filename in self.includes:
            warnings.warn("Already included: %s" % filename)
        else:
            self.includes.add(filename)
            with open(filename, 'r') as stream:
                workdir = self.workdir
                self.workdir = os.path.dirname(filename)
                self.parse(stream)
                self.workdir = workdir

    def compile(self, filename):
        self.include_file(os.path.join(self.workdir, filename))
        self.output.flush()
        self.close_output and self.output.close()
        if len(self.conditions_stack):
            raise ValueError("mismatch if/endif")


def parse_arguments(argv=None):
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument('input', nargs=1, help='input file')
    parser.add_argument('output', nargs='?', help='output file')
    parser.add_argument('-d', '--define', dest='defines', action='append', help='custom defines', default=list())
    return parser.parse_args(argv)


def main(argv=None):  # pragma: no cover
    args = parse_arguments(argv)
    if args.output:
        builder = Translator(open(args.output, 'w'), True)
    else:
        import sys
        builder = Translator(sys.stdout)

    for d, v in map(lambda x: x.split(':', 1), args.defines):
        builder.variables[d] = v

    builder.compile(args.input[0])

if __name__ == "__main__":
    main()



"""
This file is part of WSQL-SDK

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__author__ = "@bg"

from unittest import TestCase, mock
from io import BytesIO, StringIO
from warnings import catch_warnings
from wsql_sdk import codegen

try:
    from test.codegen_data import TEST_DATA
except ImportError:  # pragma: no cover
    from .codegen_data import TEST_DATA


class Dummy:
    pass


class TestCodeGen(TestCase):
    def test_validate(self):
        """ test validate the procedures """
        proc = Dummy()
        proc.name = "test"
        r1 = Dummy()
        r1.name = ""
        r1.type = "array"
        r2 = Dummy()
        r2.name = ""
        r2.type = "object"
        proc.returns = (r1, r2)
        proc.return_mod = "union"
        proc.queries = []

        self.assertRaisesRegex(ValueError, "cannot union of returns with different types", codegen.Builder.validate, proc)

    def test_parse_cmdline(self):
        """ test parse commandline arguments """
        args = codegen.parse_arguments(["-o", "build", "-l", "python3_aio", '--sep', '::', "test.sql"])
        self.assertEqual("test.sql", args.input)
        self.assertEqual("build", args.outdir)
        self.assertEqual("python3_aio", args.language)
        self.assertEqual("::", args.sep)

    def test_syntax(self):
        """ test generate result with different templates """
        for data in TEST_DATA:
            for lang in ("python3_aio", "python3"):
                args = Dummy()
                args.input = BytesIO(data["sql"])
                args.language = lang
                args.outdir = ""
                args.sep = '.'

                opened_files = dict()

                def _open_mock(fname):
                    r = opened_files[fname] = StringIO()
                    r.close = lambda: None
                    return r

                with mock.patch('builtins.open', lambda f, *a, **kw: _open_mock(f)):
                    codegen.process(args)

                filename = data.get("filename", "__init__.py")

                self.assertIn(filename, opened_files)

                code = opened_files[filename]
                code.seek(0)
                self.assertIn(data[lang], code.read())
                for n in ("constants", "exceptions"):
                    fn = n + ".py"
                    if n in data:
                        self.assertIn(fn, opened_files)
                        file = opened_files[fn]
                        file.seek(0)
                        self.assertIn(data[n], file.read())
                    else:
                        self.assertNotIn(fn, opened_files)

    def test_duplicate_fields(self):
        """test the duplicated field fire warning"""
        r1 = Dummy()
        r1.name = ""
        r1.type = "object"
        r1.fields = ("a", "b", "c")
        r2 = Dummy()
        r2.name = ""
        r2.type = "object"
        r2.fields = ("a", "c")
        proc = Dummy()
        proc.name = "test"
        proc.arguments = []
        proc.temptable = None
        proc.return_mod = "union"
        proc.returns = (r1, r2)

        with catch_warnings(record=True) as log:
            codegen.Procedure(proc.name, "", proc, True, [])
            self.assertGreater(len(log), 0)
            self.assertIn("test has duplicated fields: a, c", str(log[0]))

        r2.name = "c"
        r2.type = "object"
        r2.fields = ("q", "e")

        with catch_warnings(record=True) as log:
            codegen.Procedure(proc.name, "", proc, True, [])
            self.assertGreater(len(log), 0)
            self.assertIn("test has duplicated fields: c", str(log[0]))

        r2.name = "c"
        r2.type = "array"
        r2.fields = ("q", "e")

        with catch_warnings(record=True) as log:
            codegen.Procedure(proc.name, "", proc, True, [])
            self.assertGreater(len(log), 0)
            self.assertIn("test has duplicated fields: c", str(log[0]))



"""
This file is part of WSQL-SDK

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__author__ = "@bg"

from wsql_sdk._lang._python import *

includes_for_api = """
from wsql import Error, handle_error
from wsql.cluster import transaction
from . import exceptions"""

file_ext = ".py"
indent = "    "
doc_indent = indent
break_lines = 2


return_array = """__cursor.fetchxall()"""
return_object = return_array + "[0]"


def temporary_table(name, columns):
    """create a temporary table"""
    columns_def = ', '.join('`{0.name}` {0.type}'.format(x) for x in columns)
    column_names = ', '.join('"{0.name}"'.format(x) for x in columns)
    column_names_sql = ', '.join('`{0.name}`'.format(x) for x in columns)
    place_holders = ', '.join(["%s"] * len(columns))

    return """\
            if {0} is None:
                return
            __args = ((x.get(y, None) for y in ({1})) for x in {0})
            __cursor.execute(b"DROP TEMPORARY TABLE IF EXISTS `{0}`;")
            __cursor.execute(b"CREATE TEMPORARY TABLE `{0}`({2}) ENGINE=MEMORY;")
            __cursor.executemany(b"INSERT INTO `{0}` ({3}) VALUES ({4});", __args)"""\
        .format(name, column_names, columns_def, column_names_sql, place_holders)


def transaction_open():
    """open transaction scope"""
    return "    @transaction"


def transaction_close():
    """close transaction scope"""
    pass


def procedure_open(name, args):
    """open procedure body"""

    args = ', '.join('{0}=None'.format(x) for x in args)
    if args:
        args = ', ' + args

    return "def {0}(connection{1}):".format(name, args)


def procedure_close():
    """close procedure body"""

    return """
    try:
        return connection.execute(__query)
    except Error as e:
        raise handle_error(exceptions, e)"""


def body_open():
    """open the main logic"""
    return "    def __query(__connection):"


def body_close():
    """close the main logic"""
    pass


def cursor_open():
    """open cursor"""
    return "        with __connection.cursor() as __cursor:"


def cursor_close():
    """close cursor"""
    pass


def procedure_call(name, args):
    """call procedure"""

    args_str = ', '.join(x.name for x in args)
    if len(args) == 1:
        args_str += ","

    return '            __cursor.callproc(b"{0}", ({1}))'.format(name, args_str)



"""
This file is part of WSQL-SDK

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
__author__ = "@bg"

from collections import defaultdict, namedtuple
from functools import reduce
from pyparsing import alphanums, CaselessKeyword, Combine, delimitedList, Group, Keyword, Literal, \
    lineEnd, lineStart, MatchFirst, nestedExpr, nums, oneOf, Optional, Regex, Suppress, Word, \
    quotedString, SkipTo, Forward, White
import warnings


def _sql_identifier(expr):
    return Suppress(Optional('`')) + expr + Suppress(Optional('`'))


def _macros(expr):
    return lineStart + Suppress(Literal("#")) + expr


def _sql_comment(expr):
    return lineStart + Suppress(Literal("-- ")) + expr


def _select_hint(expr):
    return Suppress(Literal("-- >")) + expr

# keywords
_DEFINE = _macros(CaselessKeyword("define"))
_UNDEF = _macros(CaselessKeyword("undef"))
_INCLUDE = _macros(CaselessKeyword("include"))
_IF = _macros(CaselessKeyword("if"))
_ELSE = _macros(CaselessKeyword("else"))
_ENDIF = _macros(CaselessKeyword("endif"))

_AS = CaselessKeyword("AS")
_CALL = CaselessKeyword("CALL")
_COMMENT = CaselessKeyword("COMMENT")
_CREATE = CaselessKeyword("CREATE")
_DEFINER = CaselessKeyword("DEFINER")
_PROCEDURE = CaselessKeyword("PROCEDURE")
_TABLE = CaselessKeyword("TABLE")
_INTO = CaselessKeyword("INTO")
_FROM = CaselessKeyword("FROM")
_ENUM = CaselessKeyword("ENUM")
_SET = CaselessKeyword("SET")
_SELECT = CaselessKeyword("SELECT").setResultsName("command")
_DELETE = CaselessKeyword("DELETE").setResultsName("command")
_INSERT = CaselessKeyword("INSERT").setResultsName("command")
_UPDATE = CaselessKeyword("UPDATE").setResultsName("command")
IF_NOT_EXISTS = Group("IF") + CaselessKeyword("NOT") + CaselessKeyword("EXISTS")

_THROW = _sql_identifier(Keyword("__throw"))
_RETURNS = CaselessKeyword("returns")
_SQL_EOL = Literal(";")


_ID = Word(alphanums + "_.").setResultsName("name")
_ID_LIST = delimitedList(_ID, combine=False)
_VALUE = (Combine(_ID + Literal("(") + SkipTo(")", include=True)) |
          Word(alphanums + "$!#%&*+-./:<=>?@[\]^_~`") | quotedString).setResultsName("value")
_VALUE_LIST = delimitedList(_VALUE, combine=False)

_SQL_DIRECTION = oneOf('INOUT IN OUT', caseless=True)
_SQL_TYPE = Combine(Word(alphanums + '_') + Optional('(' + delimitedList(Word(nums + ' '), combine=True) + ')'))
_SQL_ID = _sql_identifier(_ID)
_SQL_ID_LIST = delimitedList(_SQL_ID, combine=False)
_SKIP_TO_END = SkipTo(_SQL_EOL)
_PARENTHESES_EXPR = Forward()
_PARENTHESES_EXPR << Combine("(" + SkipTo(")", ignore=_PARENTHESES_EXPR, include=True))
_NESTED_SELECT = Combine("(" + _SELECT + SkipTo(")", include=True, ignore=_PARENTHESES_EXPR))
_NESTED_CALL = Combine(_SQL_ID + Literal("(") + SkipTo(Literal(")"), include=True))
_SELECT_COLUMN = (_NESTED_SELECT | _NESTED_CALL | _SQL_ID.setResultsName("name")) + \
    Optional(_AS + _SQL_ID.setResultsName("alias"))

_TABLE_NAME = _SQL_ID.setResultsName("table")
_PROCEDURE_NAME = _sql_identifier(Word(alphanums + "_.:").setResultsName("name")).setResultsName('name')

_SELECT_COLUMN_LIST = delimitedList(Group(_SELECT_COLUMN), combine=False).setResultsName("columns")
_SQL_ARG = Optional(_SQL_DIRECTION, default='IN') + _SQL_ID + _SQL_TYPE
_SQL_ARGS = delimitedList(Group(_SQL_ARG), combine=False)

_RETURN_TYPE = oneOf("object array", caseless=True).setResultsName("type")

# expressions
_DEFINE_FUNCTION = _DEFINE + _ID + nestedExpr(content=_ID_LIST, ignoreExpr=None).setResultsName("args") + \
    Suppress(White()) + Regex(".*$").setResultsName("body")

_DEFINE_VAR = _DEFINE + _ID + SkipTo(lineEnd, include=True).setResultsName("value")
_UNDEFINE = _UNDEF + _ID
_INCLUDE_FILE = _INCLUDE + quotedString.setResultsName("filename")
_EXPAND_VAR = Suppress('$') + _ID
_EXPAND_FUNC = Suppress('$') + _ID + nestedExpr(content=_VALUE_LIST, ignoreExpr=None).setResultsName("args")

_IF_EXPR = _IF + SkipTo(lineEnd, include=True).setResultsName("condition")
_ELSE_EXPR = _ELSE + lineEnd
_ENDIF_EXPR = _ENDIF + lineEnd

_RETURN_HINT_EXPR = _select_hint(Optional(_ID + Suppress(Literal(":"))) + _RETURN_TYPE).setResultsName("hint")
_TEMP_TABLE_EXPR = _SQL_ID.setResultsName('table') + \
    nestedExpr(content=delimitedList(Group(_SQL_ID + _SQL_TYPE)), ignoreExpr=None).setResultsName('columns') +\
    Suppress(Literal(';'))

_PROCEDURE_COMMENT_FORMAT = Optional(_TEMP_TABLE_EXPR) + Optional(_RETURNS + oneOf("union", caseless=True).setResultsName("mode")) + lineEnd

_CREATE_PROCEDURE = _CREATE + Optional(_DEFINER + '=' + _SQL_ID) + _PROCEDURE + _PROCEDURE_NAME + \
    nestedExpr(content=_SQL_ARGS, ignoreExpr=None).setResultsName('args') + Optional(_COMMENT + quotedString.setResultsName("comment"))

_END_PROCEDURE = Regex('END\s*\$\$')

_SELECT_EXPR = _SELECT + _SELECT_COLUMN_LIST + Optional(Group(_INTO + _SQL_ID_LIST).setResultsName("into")) + \
    Optional(_FROM + _TABLE_NAME) + _SKIP_TO_END + Suppress(_SQL_EOL) + Optional(_RETURN_HINT_EXPR)

_INSERT_EXPR = _INSERT + _INTO + _TABLE_NAME + _SKIP_TO_END
_UPDATE_EXPR = _UPDATE + _TABLE_NAME + _SKIP_TO_END
_DELETE_EXPR = _DELETE + _FROM + _TABLE_NAME + _SKIP_TO_END

_THROW_EXPR = _CALL + _THROW + nestedExpr(content=_VALUE_LIST, ignoreExpr=None).setResultsName("args")

_CALL_EXPR = _CALL + _PROCEDURE_NAME + _SKIP_TO_END

_CONSTANT = _sql_comment(Keyword("CONSTANT")) + _ID + _VALUE

_CREATE_TABLE = _CREATE + _TABLE + Suppress(Optional(IF_NOT_EXISTS)) + _SQL_ID.setResultsName('name') + \
    SkipTo(";").setResultsName('body')

_DECLARE_OPTIONS = _SQL_ID.setResultsName('name') + (_ENUM | _SET).setResultsName('kind') + \
    nestedExpr(content=delimitedList(quotedString, ',', combine=False), ignoreExpr=None).setResultsName('options')


class MacrosTokenizer:
    """Preprocessor statements tokenizer"""

    _expand = \
        _EXPAND_FUNC.setResultsName("expand_function") | \
        _EXPAND_VAR.setResultsName('expand_var')

    grammar = \
        _INCLUDE_FILE.setResultsName('include') | \
        _DEFINE_FUNCTION.setResultsName('define') | \
        _DEFINE_VAR.setResultsName('define') | \
        _UNDEFINE.setResultsName('undefine') | \
        _IF_EXPR.setResultsName('if') | \
        _ELSE_EXPR.setResultsName('else') | \
        _ENDIF_EXPR.setResultsName('endif') | \
        _expand

    function_class = namedtuple('Function', ('args', 'body', 'ast'))

    def __init__(self):
        """constructor"""
        self.suppress = False
        self.conditions_stack = list()
        self.functions = dict()
        self.variables = dict()

    def reset(self):
        self.suppress = False
        self.conditions_stack.clear()
        self.functions.clear()
        self.variables.clear()

    def _expand_var(self, target):
        """expand the macros"""
        name = target.name
        if name in self.variables:
            return self.variables[name]
        return '$' + name

    def _expand_function(self, target):
        """expand the macro function"""
        macros = self.functions[target.name]
        args = (x.startswith('$') and self.variables.get(x[1:], x) or x for x in target.args[0])
        args = dict(zip(macros.args, args))
        if len(args) != len(macros.args):
            raise ValueError("invalid number of parameters for %s, expected %s" % (target.name, macros.args))

        start = 0
        buffer = ''
        for t in macros.ast:
            buffer += macros.body[start:t[1]]
            buffer += args[t[0].getName()]
            start = t[2]
        return buffer + macros.body[start:]

    def _recurisve_expand(self, value):
        """recursive expand macros"""
        buffer = ''
        start = 0
        for t in self._expand.scanString(value):
            buffer += value[start:t[1]]
            buffer += getattr(self, '_' + t[0].getName())(t[0])
            start = t[2]
        return buffer + value[start:]

    def _handle_define(self, line, token):
        """define macro function"""
        if self.suppress:
            return

        if token.args:
            args = token.args[0]
            keywords = MatchFirst([Keyword('$' + x).setResultsName(x) for x in args])
            body = self._recurisve_expand(token.body)
            macros = self.function_class(args, body, list(keywords.scanString(body)))
            if token.name in self.functions:
                warnings.warn('%d: macros %s already defined!' % (line, token.name))
            self.functions[token.name] = macros
        else:
            if token.name in self.variables:
                warnings.warn('%d: macros %s already defined!' % (line, token.name))

            value = self.variables[token.name] = self._recurisve_expand(token.value[0])
            if not token.name.startswith("_"):
                self.on_constant(token.name, value)

    def _handle_undefine(self, line, token):
        """undefine the function or variable"""
        if self.suppress:
            return

        if token.name in self.variables:
            del self.variables[token.name]
        elif token.name in self.functions:
            del self.functions[token.name]
        else:
            warnings.warn('%d: macros %s is not defined!' % (line, token.name))

    def _handle_expand_function(self, line, token):
        """handle expand macro function"""

        if self.suppress:
            return

        macros = self.functions[token.name]
        if len(macros.body) == 0:
            return

        args = dict(zip(macros.args, token.args[0]))
        if len(args) != len(macros.args):
            raise ValueError("%d: invalid number of parameters for %s, expected %s" % (line, token.name, macros.args))

        self.on_function(macros.ast, macros.body, args)

    def _handle_expand_var(self, _, token):
        """handle expand macro variable"""
        if not self.suppress:
            if token.name in self.variables:
                self.on_variable(token.name, self.variables[token.name])
            else:
                warnings.warn("Undefined variable: %s" % token.name)
                self.on_variable(token.name, "$" + token.name)

    def _handle_include(self, _, token):
        """include file"""
        if not self.suppress:
            self.on_include(token.filename.strip("'\""))

    def _handle_if(self, _, token):
        """main way of condition"""
        if not self.suppress:
            self.conditions_stack.append(self.suppress)
            self.suppress = not eval(token.condition[0], {'defined': lambda x: x in self.variables}, self.variables)

    def _handle_else(self, *_):
        """alternative way of condition"""
        self.suppress = not self.suppress

    def _handle_endif(self, line, _):
        """end of condition"""
        try:
            self.suppress = self.conditions_stack.pop()
        except IndexError:
            raise ValueError("%d: mismatch if/endif" % line) from None

    def on_constant(self, name, value):
        """callback to catch constants"""
        pass

    def on_function(self, ast, body, args):
        """callback to catch function"""
        pass

    def on_variable(self, name, value):
        """callback to catch variable"""
        pass

    def on_include(self, filename):
        """callback to catch include"""
        pass

    def nop(self, text):
        """do nothing"""
        pass

    def parse(self, stream):
        """parse the stream"""
        current = ''

        for line, text in enumerate(map(lambda x: x.rstrip('\n '), stream)):
            if not text:
                continue

            if text[-1] == '\\':
                current += text[:-1]
                continue

            current += text + '\n'
            start = 0
            tokens = self.grammar.scanString(current)
            for token in tokens:
                if not self.suppress:
                    self.nop(current[start:token[1]])
                getattr(self, '_handle_' + token[0].getName())(line, token[0])
                start = token[2]

            if not self.suppress:
                self.nop(current[start:])
            current = ''


class _Procedure:
    """Procedure abstraction"""

    argument_class = namedtuple('Argument', ('direction', 'name', 'type'))
    command_class = namedtuple('Command', ('op', 'table', 'columns'))
    temptable_class = namedtuple('TempTable', ('name', 'columns'))
    column_class = namedtuple('Column', ('name', 'type'))
    returns_class = namedtuple('Return', ('name', 'type', 'fields'))

    def __init__(self, name, arguments, comment):
        self.name = name
        self.arguments = [self.argument_class(*x) for x in arguments if x]
        self.queries = list()
        self.modifiers = list()
        self.children = list()
        self.read_only = None
        self.errors = set()
        self.temptable = None
        self.returns = list()
        self.return_mod = None

        if comment:
            try:
                hint = _PROCEDURE_COMMENT_FORMAT.parseString(comment.strip("'\""))
                self.return_mod = hint.mode
                if hint.table:
                    self.temptable = self.temptable_class(hint.name, tuple(self.column_class(*x) for x in hint.columns[0]))
            except Exception as e:
                raise ValueError('SyntaxError: procedure %s, comment %s: %s' % (name, comment, e))

    def add_read_command(self, *args):
        """handle new read command"""
        self.queries.append(self.command_class(*args))

    def add_write_command(self, *args):
        """handle new write command"""
        self.modifiers.append(self.command_class(*args))
        self.read_only = False

    def add_return(self, *args):
        self.returns.append(self.returns_class(*args))

    def __repr__(self):
        return self.name


class SQLTokenizer:
    """The sql statement tokenizer"""
    procedure_class = _Procedure

    def __init__(self):
        self._grammar = \
            _CREATE_PROCEDURE.copy().setParseAction(self.on_begin_procedure) | \
            _END_PROCEDURE.copy().setParseAction(self.on_end_procedure) | \
            _SELECT_EXPR.copy().setParseAction(self.on_select) | \
            _INSERT_EXPR.copy().setParseAction(self.on_insert) | \
            _UPDATE_EXPR.copy().setParseAction(self.on_update) | \
            _DELETE_EXPR.copy().setParseAction(self.on_delete) | \
            _THROW_EXPR.copy().setParseAction(self.on_error) | \
            _CALL_EXPR.copy().setParseAction(self.on_call) | \
            _CONSTANT.copy().setParseAction(self.on_constant) | \
            _CREATE_TABLE.copy().setParseAction(self.on_table)

        self._procedures = dict()
        self._constants = list()
        self._structures = dict()
        self._current = None

    @staticmethod
    def _column_name(column):
        if column.alias:
            return column.alias[0]
        if column.name:
            return column.name[0].rpartition('.')[-1]
        return str(column)

    def reset(self):
        self._procedures.clear()
        self._current = None

    def on_begin_procedure(self, tokens):
        """catch the begin of procedure"""
        self._current = self.procedure_class(tokens.name[0], tokens.args[0], tokens.comment)
        if self._current.name in self._procedures:
            raise ValueError('The procedure %s already defined!' % self._current)
        self._procedures[self._current.name] = self._current

    def on_end_procedure(self, _):
        """catch the end of procedure"""
        self._current = None

    def on_select(self, tokens):
        """catch the select statement"""
        if self._current and not tokens.into:
            columns = tuple(self._column_name(x) for x in tokens.columns)
            self._current.add_read_command(tokens.op, tokens.table and tokens.table[0], columns)
            return_hint = tokens.hint
            if return_hint:
                name = return_hint.name
                rtype = return_hint.type
            else:
                name = ""
                rtype = ""

            self._current.add_return(name, rtype or "object", columns)

    def on_insert(self, tokens):
        """catch the modify statement"""
        if self._current:
            self._current.add_write_command(tokens.op, tokens.table, [])

    on_update = on_insert
    on_delete = on_insert

    def on_error(self, tokens):
        """catch the raising of the exception"""
        if self._current:
            self._current.errors.add(tokens.args[0][0].strip("'\""))

    def on_constant(self, token):
        """catch constants"""
        self._constants.append((token.name, token.value))

    def on_table(self, tokens):
        """catch the table"""
        table_name = tokens.name.name
        structures = defaultdict(list)
        for t in _DECLARE_OPTIONS.scanString(tokens.body):
            t = t[0]
            structures[t.kind].append((t.name.name, sorted(x.strip('"\'') for x in t.options[0])))

        if len(structures) > 0:
            self._structures[table_name] = structures

    def on_call(self, tokens):
        """catch the procedure call"""
        if self._current:
            self._current.children.append(tokens[1])

    def parse(self, text):
        """
        parse the input text
        :param text: the sql statements
        """
        for _ in self._grammar.scanString(text):
            pass

    def constants(self):
        """
        :return: the list of constants
        """
        return self._constants

    def structures(self, name):
        """
        :return: the list of enums
        """
        if name in self._structures:
            return self._structures[name]

    def procedures(self):
        """
        enumerate all of procedures
        :return: the enumerator of all procedures
        """
        return self._procedures.values()

    def errors(self, procedure=None):
        """
        get procedure errors recursively
        :param procedure: optional, start procedure
        :return the set of errors
        :rtype set
        """
        if procedure is None:
            return reduce(lambda x, y: x | y.errors, self._procedures.values(), set())

        errors = procedure.errors.copy()
        for child in set(procedure.children):
            errors |= self.errors(self._procedures[child])
        return errors

    def queries(self, procedure):
        """
        recursively return all queries that performed in specified procedure
        :param procedure: the procedure object
        :return: the list of all queries
        """
        queries = procedure.queries[:]
        for child in (self._procedures[x] for x in procedure.children):
            queries.extend(self.queries(child))
        return queries

    def is_read_only(self, procedure):
        """recursively determine that procedure read-only or not"""
        if procedure.read_only is not None:
            return procedure.read_only

        for child in (self._procedures[x] for x in procedure.children):
            if self.is_read_only(child) is False:
                procedure.read_only = False
                return False

        procedure.read_only = True
        return True



#!/usr/bin/env python

from setuptools import setup

try:
    from sphinx.setup_command import BuildDoc as _BuildDoc

    class BuildDoc(_BuildDoc):
        def finalize_options(self):
            super().finalize_options()
            if not self.project:
                self.project = self.distribution.name
            if not self.version:
                self.version = self.distribution.version

except ImportError:
    BuildDoc = None


def readme():
    with open('README') as r:
        return r.read()


setup(
    name="wsql_sdk",
    version="0.3.7",
    description='The chain of tools, that to make work with SQL easier',
    packages=["wsql_sdk", "wsql_sdk._lang"],
    requires=["pyparsing"],
    author="Bulat Gaifullin",
    author_email='gaifullinbf@gmail.com',
    maintainer='Bulat Gaifullin',
    maintainer_email='gaifullinbf@gmail.com',
    url='https://github.com/WebSQL/sdk',
    license='MIT',
    long_description=readme(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: OS Independent",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
    ],
    entry_points={
        'console_scripts': [
            'wsql-codegen=wsql_sdk.codegen:main',
            'wsql-trans=wsql_sdk.translator:main',
        ],
    }
)



"""
This file is part of WSQL-SDK

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__author__ = "@bg"

from unittest import TestCase
from wsql_sdk import grammar


class TestInterpreter(TestCase):
    def test_ids(self):
        """test the sql id syntax"""
        self.assertEqual("test", grammar._ID.parseString("test").name)
        self.assertEqual("test.test", grammar._ID.parseString("test.test").name)
        self.assertEqual("test_123", grammar._ID.parseString("test_123").name)

        self.assertEqual("test", grammar._SQL_ID.parseString("test")[0])
        self.assertEqual("test.test", grammar._SQL_ID.parseString("`test.test`")[0])
        self.assertEqual("test_123", grammar._SQL_ID.parseString("test_123")[0])

        self.assertEqual(["test", "test_123", "test.test"],
                         grammar._ID_LIST.parseString("test, test_123, test.test").asList())
        self.assertEqual(["test", "test_123", "test.test"],
                         grammar._SQL_ID_LIST.parseString("test, `test_123`, `test.test`").asList())

    def test_values(self):
        """test sql value syntax"""
        self.assertEqual("test", grammar._VALUE.parseString("test")[0])
        self.assertEqual("test(*)", grammar._VALUE.parseString("test(*)")[0])
        self.assertEqual("test(123)", grammar._VALUE.parseString("test(123)")[0])
        self.assertEqual("123", grammar._VALUE.parseString("123")[0])
        self.assertEqual("*", grammar._VALUE.parseString("*")[0])
        self.assertEqual("'test test'", grammar._VALUE.parseString("'test test'")[0])
        self.assertEqual("'test, test'", grammar._VALUE.parseString("'test, test'")[0])

        self.assertEqual(["test", "test(*)", "123", "'test, test'"],
                         grammar._VALUE_LIST.parseString("test, test(*), 123, 'test, test'").asList())

    def test_sql_argument(self):
        """the sql argument syntax"""
        self.assertEqual(["IN", "test", "VarChar(255)"],
                         grammar._SQL_ARG.parseString("In test VarChar(255)").asList())
        self.assertEqual(["OUT", "test.test", "INT"],
                         grammar._SQL_ARG.parseString("Out `test.test` INT").asList())
        self.assertEqual(["IN", "test_test", "BOOL(1)"],
                         grammar._SQL_ARG.parseString("test_test BOOL(1)").asList())

    def test_parentheses_expr(self):
        """"the the parentheses syntax"""
        self.assertEqual("(a(b(c(d))))", grammar._PARENTHESES_EXPR.parseString("(a(b(c(d))))")[0])

    def test_select_column(self):
        """test the select column syntax"""
        self.assertEqual("(SELECT * FROM A WHERE id > 0)",
                         grammar._NESTED_SELECT.parseString("(SELECT * FROM A WHERE id > 0)")[0])
        self.assertEqual("(SELECT COUNT(*) FROM A WHERE id > 0)",
                         grammar._NESTED_SELECT.parseString("(SELECT COUNT(*) FROM A WHERE id > 0)")[0])
        self.assertEqual("EXISTS(SELECT * FROM A WHERE id > 0)",
                         grammar._NESTED_CALL.parseString("EXISTS(SELECT * FROM A WHERE id > 0)")[0])
        self.assertEqual("MAX(*)",
                         grammar._NESTED_CALL.parseString("MAX(*)")[0])
        self.assertEqual("count",
                         grammar._SELECT_COLUMN.parseString("`count`").name[0])

        self.assertEqual(["count", "max", "id"],
                         [x[-1] for x in
                          grammar._SELECT_COLUMN_LIST.parseString(
                              "(SELECT COUNT(*) FROM A) AS `count`, MAX(*) AS `max`, id").columns])

    def test_define_variable(self):
        """test the define variable syntax"""
        self.assertEqual(['define', 'test', '"test"'],
                         grammar._DEFINE_VAR.parseString("#define test \"test\"").asList())

        self.assertEqual(['define', 'test', "f(w,x)"],
                         grammar._DEFINE_VAR.parseString("#define test f(w,x)").asList())

        self.assertEqual(['define', 'test', '"test1 test2"'],
                         grammar._DEFINE_VAR.parseString("#define test \"test1 test2\"").asList())

    def test_define_function(self):
        """test the define function syntax"""
        self.assertEqual(['define', 'test', ['a1', 'a2', 'a3'], 'f($a1, $a2, $a3)'],
                         grammar._DEFINE_FUNCTION.parseString("#define test(a1,a2,a3) f($a1, $a2, $a3)").asList())

    def test_undefine(self):
        """test the undefine syntax"""
        self.assertEqual(['undef', 'test'],
                         grammar._UNDEFINE.parseString("#undef test").asList())

    def test_expand_var(self):
        """test expand variable syntax"""
        self.assertEqual("test",
                         grammar._EXPAND_VAR.parseString("$test").name)

    def test_expand_func(self):
        """test expand  function syntax"""
        self.assertEqual(["test", ["a1", "\"a b\"", "f(w,x)"]],
                         grammar._EXPAND_FUNC.parseString("$test(a1, \"a b\", f(w,x))").asList())

    def test_include(self):
        """test the include syntax"""
        self.assertEqual(["include", "\"test.sql\""],
                         grammar._INCLUDE_FILE.parseString("#include \"test.sql\"").asList())

    def test_return_hint(self):
        """test return hint syntax"""
        self.assertEqual(["test", "object"], grammar._RETURN_HINT_EXPR.parseString("-- > test:object").asList())
        self.assertEqual(["object"], grammar._RETURN_HINT_EXPR.parseString("-- > object").asList())

    def test_temp_table(self):
        """test temp table syntax"""
        self.assertEqual(["test", [["c1", "INT"], ["c2", "VARCHAR(200)"]]],
                         grammar._TEMP_TABLE_EXPR.parseString("test(c1 INT, c2 VARCHAR(200));").asList())

    def test_meta(self):
        """the the procedure meta syntax"""
        self.assertEqual(["test", [["c1", "INT"], ["c2", "VARCHAR(200)"]], "returns", "union"],
                         grammar._PROCEDURE_COMMENT_FORMAT.parseString("test(c1 INT, c2 VARCHAR(200)); returns union").asList())
        self.assertEqual(["returns", "union"],
                         grammar._PROCEDURE_COMMENT_FORMAT.parseString("returns union").asList())

    def test_create_procedure(self):
        """test create procedure syntax"""
        self.assertEqual(["CREATE", "DEFINER", "=", "test", "PROCEDURE", "test",
                          [["IN", "a1", "INT"], ["IN", "a_2", "VARCHAR(255)"], ["OUT", "a.b", "BOOL"]], "COMMENT", "\"returns union\""],
                         grammar._CREATE_PROCEDURE.parseString(
                             "CREATE DEFINER = `test` PROCEDURE test(a1 INT,IN a_2 VARCHAR(255), OUT `a.b` BOOL) COMMENT \"returns union\"").asList())

    def test_select(self):
        """test the select statement syntax"""
        self.assertEqual(['SELECT',
                         ['MAX(*)', 'AS', 'a'], ['(SELECT 1 FROM Q)', 'AS', 'b'], ['c', 'AS', 'c.a'],
                         'FROM', 'T', ''],
                         grammar._SELECT_EXPR.parseString("SELECT MAX(*) AS a, (SELECT 1 FROM Q) AS b,"
                                                          "c AS `c.a` FROM T;").asList())
        self.assertEqual(['SELECT',
                          ['MAX(*)', 'AS', 'a'], ['(SELECT 1 FROM Q)', 'AS', 'b'], ['c', 'AS', 'c.a'],
                          ['INTO', 'a', 'b', 'c'], 'FROM', 'T', '', 'object'],
                         grammar._SELECT_EXPR.parseString("SELECT MAX(*) AS a, (SELECT 1 FROM Q) AS b,"
                                                          "c AS `c.a` INTO a,b,c FROM T; -- > object").asList())

    def test_insert(self):
        """test the insert statement syntax"""
        self.assertEqual(['INSERT', 'INTO', 'test', '(a,b) VALUES (1,2)'],
                         grammar._INSERT_EXPR.parseString("INSERT INTO test (a,b) VALUES (1,2);").asList())

    def test_update(self):
        """test the update statement syntax"""
        self.assertEqual(['UPDATE', 'test', 'set a=1'],
                         grammar._UPDATE_EXPR.parseString("UPDATE test set a=1;").asList())

    def test_delete(self):
        """test the delete statement syntax"""
        self.assertEqual(['DELETE', 'FROM', 'test', ''],
                         grammar._DELETE_EXPR.parseString("DELETE FROM test;").asList())

    def test_throw(self):
        """test throw syntax"""
        self.assertEqual(['CALL', '__throw', ["'test'", "'this is test message'"]],
                         grammar._THROW_EXPR.parseString("CALL __throw('test', 'this is test message');").asList())

    def test_call(self):
        """test procedure call syntax"""
        self.assertEqual(['CALL', 'proc1', "(1, 2)"],
                         grammar._CALL_EXPR.parseString("CALL proc1(1, 2);").asList())

    def test_create_table(self):
        """test create table syntax"""
        self.assertEqual(
            ['CREATE', 'TABLE', 'T1', '(\nc1 ENUM("a", "b", "c"), c2 SET("0", "1", "2")\n)'],
            grammar._CREATE_TABLE.parseString(
                'CREATE TABLE IF NOT EXISTS `T1`(\nc1 ENUM("a", "b", "c"), c2 SET("0", "1", "2")\n);'
            ).asList()
        )

    def test_column_options(self):
        """test column options declare syntax"""
        found = list(grammar._DECLARE_OPTIONS.scanString('(c1 ENUM("a", "b", "c"), c2 SET("0", "1", "2"));'))
        self.assertEqual(2, len(found))
        self.assertEqual('c1', found[0][0].name[0])
        self.assertEqual('ENUM', found[0][0].kind)
        self.assertEqual(['"a"', '"b"', '"c"'], found[0][0].options.asList()[0])
        self.assertEqual('c2', found[1][0].name[0])
        self.assertEqual('SET', found[1][0].kind)
        self.assertEqual(['"0"', '"1"', '"2"'], found[1][0].options.asList()[0])

    def test_sql_type(self):
        found = list(grammar._SQL_TYPE.scanString("BINARY(16), INT, DECIMAL(32, 16)"))
        self.assertEqual(3, len(found))
        self.assertEqual("BINARY(16)", found[0][0][0])
        self.assertEqual("INT", found[1][0][0])
        self.assertEqual("DECIMAL(32, 16)", found[2][0][0])
